@@ -28,6 +28,7 @@
 #include <IOKit/hidsystem/IOHIDTypes.h>
 #include <IOKit/hidsystem/IOHIKeyboard.h>
 #include "IOHIDDevice.h"
+#include "IOHIDConsumer.h"
 
 enum {
     kUSB_LEFT_CONTROL_BIT = 0x01,
@@ -42,7 +43,6 @@ enum {
 };
 
 enum {
-    kUSB_LOWSPEED_MAXPACKET = 8,
     kUSB_CAPSLOCKLED_SET = 2,
     kUSB_NUMLOCKLED_SET = 1
 };
@@ -63,22 +63,51 @@ enum {
     kgestUSBAndyJISKbd        = 206,      /* (0xCE) Andy USB Keyboard Japanese (JIS) Keyboard */
     kprodUSBAndyJISKbd	      = 0x206,
 
-    kgestQ6ANSIKbd	      = 31,      /* (0x31) Apple Q6 Keyboard Domestic (ANSI) Keyboard */
+    kgestQ6ANSIKbd	      = 31,      /* (031) Apple Q6 Keyboard Domestic (ANSI) Keyboard */
     kprodQ6ANSIKbd	      = 0x208,
-    kgestQ6ISOKbd	      = 32,      /* (0x32) Apple Q6 Keyboard International (ISO) Keyboard */
+    kgestQ6ISOKbd	      = 32,      /* (32) Apple Q6 Keyboard International (ISO) Keyboard */
     kprodQ6ISOKbd	      = 0x209,
-    kgestQ6JISKbd	      = 33,      /* (0x33) Apple Q6 Keyboard Japanese (JIS) Keyboard */
+    kgestQ6JISKbd	      = 33,      /* (33) Apple Q6 Keyboard Japanese (JIS) Keyboard */
     kprodQ6JISKbd	      = 0x20a,
     
-    kgestQ30ANSIKbd	      = 34,      /* (0x34) Apple Q30 Keyboard Domestic (ANSI) Keyboard */
+    kgestQ30ANSIKbd	      = 34,      /* (34) Apple Q30 Keyboard Domestic (ANSI) Keyboard */
     kprodQ30ANSIKbd	      = 0x20b,
-    kgestQ30ISOKbd	      = 35,      /* (0x35) Apple Q30 Keyboard International (ISO) Keyboard */
+    kgestQ30ISOKbd	      = 35,      /* (35) Apple Q30 Keyboard International (ISO) Keyboard */
     kprodQ30ISOKbd	      = 0x20c,
-    kgestQ30JISKbd	      = 36,      /* (0x36) Apple Q30 Keyboard Japanese (JIS) Keyboard */
-    kprodQ30JISKbd	      = 0x20d    
+    kgestQ30JISKbd	      = 36,      /* (36) Apple Q30 Keyboard Japanese (JIS) Keyboard */
+    kprodQ30JISKbd	      = 0x20d,
+    
+    kgestFountainANSIKbd      = 37,      /* (37) Apple Fountain Keyboard Domestic (ANSI) Keyboard */
+    kprodFountainANSIKbd      = 0x20e,
+    kgestFountainISOKbd	      = 38,      /* (38) Apple Fountain Keyboard International (ISO) Keyboard */
+    kprodFountainISOKbd	      = 0x20f,
+    kgestFountainJISKbd	      = 39,      /* (39) Apple Fountain Keyboard Japanese (JIS) Keyboard */
+    kprodFountainJISKbd	      = 0x210,
+
+    kgestSantaANSIKbd	      = 37,      /* (37) Apple Santa Keyboard Domestic (ANSI) Keyboard */
+    kprodSantaANSIKbd         = 0x211,
+    kgestSantaISOKbd	      = 38,      /* (38) Apple Santa Keyboard International (ISO) Keyboard */
+    kprodSantaISOKbd	      = 0x212,
+    kgestSantaJISKbd	      = 39,      /* (39) Apple Santa Keyboard Japanese (JIS) Keyboard */
+    kprodSantaJISKbd	      = 0x213,
+            
 
 };
 
+enum {
+    kSecondaryKeyFnSpecial	= 0x01,
+    kSecondaryKeyFnNonSpecial	= 0x02,
+    kSecondaryKeyNumPad		= 0x04
+};
+
+typedef struct _SecondaryKey {
+    UInt8	bits;
+    UInt8	swapping;
+    UInt8	specialKey;
+    UInt8	fnUsage;
+    UInt8	numPadUsage;
+} SecondaryKey;
+
 #define ADB_CONVERTER_LEN       0xff + 1   //length of array def_usb_2_adb_keymap[]
 
 class IOHIDKeyboard : public IOHIKeyboard
@@ -86,10 +115,16 @@ class IOHIDKeyboard : public IOHIKeyboard
     OSDeclareDefaultStructors(IOHIDKeyboard)
 
     IOHIDDevice	*		_provider;
+    IOHIDConsumer *		_consumer;
+    IONotifier *		_publishNotify;
+    
+    IORecursiveLock *		_keyboardLock;
 
     // Values from the provider
-    UInt16			_productID;
-    UInt16			_vendorID;
+    OSNumber*			_productID;
+    OSNumber*			_vendorID;
+    OSNumber*			_locationID;
+    OSString *			_transport;
         
     // LED Specific Members
     IOHIDElementCookie  	_ledCookies[2];
@@ -110,24 +145,39 @@ class IOHIDKeyboard : public IOHIKeyboard
     UInt32 *			_oldArraySelectors;
     UInt8                       _usb_2_adb_keymap[ADB_CONVERTER_LEN + 1];
     
+    // FN Key Member
+    UInt32 *			_fKeyValuePtr;
+    UInt32			_lastFKeyValue;
+    UInt8			_fKeyMode;
+    bool			_stickyKeysOn;
+    SecondaryKey    		_secondaryKeys[255];
+    
     // *** PRIVATE HELPER METHODS ***
     void   			Set_LED_States(UInt8);
     UInt32 			handlerID();
-    bool			findDesiredElements(OSArray *elements);
+    bool			determineKeyboard(IOHIDDevice *owner);
+    bool			findDesiredElements(OSArray *elements, IOHIDDevice *owner);
+    void			findSecondaryKeys();
+    bool			filterSecondaryNumPadKey(int * usage, bool down, AbsoluteTime ts);
+    bool			filterSecondaryFnSpecialKey(int * usage, bool down, AbsoluteTime ts);
+    bool			filterSecondaryFnNonSpecialKey(int * usage, bool down, AbsoluteTime ts);
+
     // *** END PRIVATE HELPER METHODS ***
     
     // static methods for callbacks, the command gate, new threads, etc.
-    static void 		AsyncLED (OSObject *target);
-
+    static void 		_asyncLED (OSObject *target);
+    static bool 		_publishNotificationHandler(void * target, void * ref, IOService * newService );
+                                
 public:    
     // Allocator
-    static IOHIDKeyboard * 	Keyboard(OSArray *elements);
+    static IOHIDKeyboard * 	Keyboard(OSArray *elements, IOHIDDevice *owner);
     
     // IOService methods
     virtual bool		init(OSDictionary * properties = 0);
     virtual bool		start(IOService * provider);
     virtual void 		stop(IOService *  provider);
     virtual void 		free();
+    virtual bool 		matchPropertyTable(OSDictionary * table, SInt32 * score);    
 
     virtual void 		handleReport();
 
@@ -140,8 +190,8 @@ public:
     const unsigned char * 	defaultKeymapOfLength(UInt32 * length);
     void 			setAlphaLockFeedback(bool LED_state);
     void 			setNumLockFeedback(bool LED_state);
-    bool 			doesKeyLock(unsigned key);
     unsigned 			getLEDStatus();
+    IOReturn 			setParamProperties( OSDictionary * dict );
 
 };
 
